// Author: Alexander S. Gerasimov

#ifndef BIN_SEARCH_TREE_MAP_H
#define BIN_SEARCH_TREE_MAP_H

// NOTE: This implementation of an ordinary binary search tree should be
// replaced with your implementation of an AVL tree!

// This template represents a binary search tree with data of a type T
// in its nodes. Elements of the type T are ordered according to the
// operator <. Two elements x and y of the type T are considered equal iff
// neither x<y nor y<x. There are no two different nodes with the same data
// in such a binary search tree.
template<typename T> class BinSearchTree {
private:
    struct Node {
        T data;
        Node* left;
        Node* right;
        Node(const T& d) : data(d), left(0), right(0) {}
    };

    Node* root;

    void destruct(Node* t);

    // Prohibit initialization and assignment
    BinSearchTree(const BinSearchTree&);
    BinSearchTree& operator=(const BinSearchTree&);

public:
    BinSearchTree() : root(0) {}
    virtual ~BinSearchTree();

    // If no data in a node of this tree are equal to d, then adds a leaf
    // with the data d to this tree; otherwise sets the data d to the node
    // that contains data equal to d.
    virtual void add(const T& d);

    // If no data in a node of this tree are equal to d, then returns 0;
    // otherwise returns the pointer to the data in a node of this tree,
    // where the data are equal to d.
    T* search(const T& d) const;
};


template<typename T> BinSearchTree<T>::~BinSearchTree() {
    destruct(root);
    root = 0;
}

template<typename T> void BinSearchTree<T>::destruct(Node* t) {
    if (t != 0) {
        destruct(t->left);
        destruct(t->right);
        delete t;
    }
}

template<typename T> T* BinSearchTree<T>::search(const T& d) const {
    Node* p = root;
    while (p) {
        if (p->data < d) {
            p = p->right;
        } else if (d < p->data) {
            p = p->left;
        } else { // p->data and d are equal
            return &p->data;
        }
    }
    return 0;
}

template<typename T> void BinSearchTree<T>::add(const T& d) {
    Node** p = &root;
    while (*p) {
        if ((*p)->data < d) {
            p = &(*p)->right;
        } else if (d < (*p)->data) {
            p = &(*p)->left;
        } else { // (*p)->data and d are equal
            (*p)->data = d;
            return;
        }
    }
    *p = new Node(d);
}


// This template represents a map of keys (of a type K) to values (of a type V).
// Keys are ordered according to the operator <.
// Two keys x and y are considered equal iff neither x<y nor y<x.
template<typename K, typename V> class Map {
private:
    struct Pair {
        K k;
        V v;
        Pair(const K& key, const V& val) : k(key), v(val) {};
        bool operator<(const Pair& p) const {
            return k < p.k;
        }
    };

    BinSearchTree<Pair> tree;

public:
    void put(const K& k, const V& v) { tree.add(Pair(k, v)); }

    V* get(const K& k) const {
        Pair* p = tree.search(Pair(k, V()));
        return (p ? &p->v : 0);
    }
};

#endif
